"""
Support code for controlling a multiphase inverter.

Victron MultiPlus only, for now.

"""

# pylint: disable=too-many-lines

from __future__ import annotations

import anyio
import logging
from contextlib import asynccontextmanager
from datetime import datetime

from asyncdbus import DBusError
from asyncdbus.service import method

from moat.ems.victron.dbus import Dbus
from moat.ems.victron.dbus.monitor import DbusMonitor
from moat.ems.victron.dbus.utils import (
    CtxObj,
    DbusInterface,
    DbusName,
    unwrap_dbus_dict,
    unwrap_dbus_value,
    wrap_dbus_dict,
)
from moat.util.times import time_until

from ._util import balance

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from asyncdbus import DbusType

    from typing import Any

__all__ = ["BusVars", "InvControl", "InvInterface", "InvModeBase"]

logger = logging.getLogger(__name__)

_dummy = {"code": None, "whenToLog": "configChange", "accessLevel": None}

__path__ = __import__("pkgutil").extend_path(__path__, __name__)


class BusVars(CtxObj):
    """
    A wrapper that creates attributes as per the class's VARS and VARS_RO arrays.

    Both are mappings:  dbusname => { attrname => path }

    The attributes registered via VARS are DbusItemImport objects.
    Attributes registered via VARS_RO are read from the bus, and set directly.
    """

    VARS = {}
    VARS_RO = {}

    _intf: Dbus = None

    def __init__(self, bus):
        self._bus = bus

    @asynccontextmanager
    async def _ctx(self):
        async with Dbus(self._bus) as self._intf:
            for k, v in self.VARS_RO.items():
                for n, p in v.items():
                    setattr(
                        self,
                        n,
                        (await self._intf.importer(k, p, createsignal=False)).value,
                    )
            for k, v in self.VARS.items():
                for n, p in v.items():
                    setattr(self, n, await self._intf.importer(k, p))
            yield self

    @property
    def bus(self):
        "Accessor for _bus"
        return self._bus

    @property
    def intf(self):
        "Accessor for _intf"
        return self._intf


class InvInterface(DbusInterface):
    """A Dbus interface for inverter control"""

    # pylint: disable=protected-access

    def __init__(self, ctrl):
        self.ctrl = ctrl
        super().__init__(ctrl.bus, "/Control", "inv")

    @method()
    async def GetModes(self) -> DbusType["as"]:  # noqa:F722
        """
        Return a list of available methods.
        name => (ident#, descr)
        """
        return [m._name for m in InvControl.MODES.values()]  # noqa:SLF001

    @method()
    async def GetModeInfo(self, mode: DbusType["s"]) -> DbusType["a{ss}"]:  # noqa:F722,UP037,F821
        """
        Return information on a specific method
        """
        m = InvControl.MODES[mode]
        return m._doc  # noqa:SLF001

    @method()
    async def SetMode(self, mode: DbusType["s"], args: DbusType["a{sv}"]) -> DbusType["b"]:  # noqa:F722,UP037,F821
        """
        Change the inverter mode, set parameters
        """
        return await self.ctrl.change_mode(mode, unwrap_dbus_dict(args))

    @method()
    async def SetModeParam(self, param: DbusType["s"], value: DbusType["v"]) -> DbusType["b"]:  # noqa:UP037,F821
        """
        Set a specific parameter
        """
        return await self.ctrl.change_mode_param(param, unwrap_dbus_value(value))

    @method()
    async def GetState(self) -> DbusType["a{sv}"]:  # noqa:F722
        """
        Return the state of the current inverter controller.
        """
        return wrap_dbus_dict(self.ctrl.get_state())


class InvControl(BusVars):
    """
    This is the main controller for an inverter.

    It can operate in various modes. Call `change_mode` to switch between them.
    There's a mandatory 30 second delay so that things can settle down somewhat.

    Configurable parameters ("system" group in ``inv*.cfg``):
    """

    # pylint: disable=too-many-public-methods

    #
    # Conventions used in this code:
    #
    # Max and min values are unsigned.
    #
    # All power values (except for solar input) refer to AC. All current values refer to DC.
    #
    # The sum of all power/current values is zero by definition.
    # Positive == power/current goes to your home  / DC bus.

    # when our algorithms say "go from X to Y" we only go partways towards Y,
    # because otherwise fun nonlinear effects (solar output adapts, battery
    # voltage changes due to internal resistance, …) cause the system to oscillate.
    f_step = 0.35
    # but if the delta is smaller than this, just set it.
    # This is also used as the max "stable" change between values, i.e. assume that
    # the system has mostly settled down if the charger/inverter load changes
    # by less than this
    p_step = 100
    # if the SoC is more than f_delta away from 0/1, don't bother either
    f_delta = 0.2

    solar_p = 0  # current solar power yield

    top_off = False  # go to the battery voltage limit?
    umax_diff = 0.5  # distance to max voltage, when not topping off
    umin_diff = 0.5  # distance to min voltage

    pg_min = -12000  # watt we may send to the grid
    pg_max = 12000  # watt we may take from the grid
    inv_eff = 0.9  # inverter's typical efficiency
    p_per_phase = 4500  # inverter's max load per phase
    # TODO collect long term deltas

    # protect battery against excessive discharge if PV current should suddenly
    # fall off due to clouds. We assume that it'll not drop more than 60%.
    pv_margin = 0.4
    # try to keep the max current from the solar chargers this many amps above the
    # current amperage so that if solar power increases the system can notice and adapt
    pv_delta = 30

    #
    cap_scale = 4
    #
    # Approximate internal resistance of the battery pack.
    # TODO should be autodetectable (dU/dI)
    r_int = 0.01
    # Per-phase variables
    # p_set_
    #   Multi, /Hub4/L{i}/AcPowerSetpoint
    #   the power we want this Multi to get/emit. Positive.
    # p_cur_
    #   System, /Ac/ActiveIn/L{i}/Power
    #   The power actually going to/from the grid. Positive.
    # p_run_
    #   Multi, /Ac/ActiveIn/L{i}/P
    #   The power the Multi is feeding us. Negative.
    # p_cons_
    #   System, /Ac/Consumption/L{i}/Power
    #   The power other consumers are taking from the bus. Negative.
    #

    # distkv
    _dkv = None
    _dkv_evt = None

    _mode = None
    MODES = {}

    # Declared in VARS and VARS_RO. Here due to pylint.
    acc_vebus = None
    s_battery = None
    n_phase = None
    _u_dc = None
    _i_batt = None
    _i_inv = None
    _i_pv = None
    _batt_soc = None
    _p_cons1 = None
    _p_cons2 = None
    _p_cons3 = None
    _p_grid1 = None
    _p_grid2 = None
    _p_grid3 = None

    p_grid_: list = None
    p_cons_: list = None
    p_cur_: list = None
    load: list = None
    u_min = None
    u_max = None
    _ib_chg = None
    _ib_dis = None
    _ok_chg = None
    _ok_dis = None
    b_cap = None
    p_set_ = None
    p_run_ = None
    _p_inv = None
    _mode_task = None
    _mode_task_stopped = None
    dest_p = None
    last_p = None
    distkv_prefix = None
    _bms_intf = None
    _batt_intf = None
    _change_mode_evt = None
    _ctrl = None
    _srv = None
    _tg = None
    step = None

    @classmethod
    def register(cls, target):
        """Register an inverter control method"""
        # pylint: disable=protected-access

        if target._name in cls.MODES:  # noqa:SLF001
            raise RuntimeError(f"Mode {target._mode} already known: {cls.MODES[target._mode]}")  # noqa:SLF001
        cls.MODES[target._name] = target  # noqa:SLF001
        return target

    MON = {
        "com.victronenergy.solarcharger": {
            "/Yield/Power": _dummy,
            "/CustomName": _dummy,
        },
    }

    VARS = {
        "com.victronenergy.system": dict(
            acc_vebus="/VebusService",
            _i_pv="/Dc/Pv/Current",
            _u_dc="/Dc/Battery/Voltage",
            _i_batt="/Dc/Battery/Current",
            _i_inv="/Dc/Vebus/Current",
            _batt_soc="/Dc/Battery/Soc",
            _p_cons1="/Ac/Consumption/L1/Power",
            _p_cons2="/Ac/Consumption/L2/Power",
            _p_cons3="/Ac/Consumption/L3/Power",
            _p_grid1="/Ac/Grid/L1/Power",
            _p_grid2="/Ac/Grid/L2/Power",
            _p_grid3="/Ac/Grid/L3/Power",
            s_battery="/Dc/Battery/BatteryService",
        ),
    }
    VARS_RO = {
        "com.victronenergy.system": dict(
            n_phase="/Ac/ActiveIn/NumberOfPhases",
        ),
    }

    def __init__(self, bus, cfg):
        super().__init__(bus)
        self.cfg = cfg
        self.op = cfg.get("op", {})

        for k, v in cfg.get("system", {}).items():
            try:
                vv = getattr(self, k)
                if not isinstance(vv, (type(None), int, float, str)):
                    raise TypeError(vv)
            except AttributeError:
                logger.error("System param unknown: %r", k)
            except RuntimeError:
                logger.error("Not a system param: %r", k)
            else:
                setattr(self, k, v)

        self._trigger = anyio.Event()
        self._dkv_evt = anyio.Event()
        self.clear_state()

    def clear_state(self):
        "clear state log"
        self._state = {}

    def get_state(self):
        "return current state log"
        return self._state

    def set_state(self, k, v):
        "update state log"
        self._state[k] = v

    @asynccontextmanager
    async def _ctx(self):
        async with super()._ctx():
            self.p_grid_ = []
            self.p_cons_ = []
            self.p_cur_ = []

            for i in range(1, self.n_phase + 1):
                self.p_grid_.append(
                    await self.intf.importer("com.victronenergy.system", f"/Ac/Grid/L{i}/Power"),
                )
                self.p_cons_.append(
                    await self.intf.importer(
                        "com.victronenergy.system",
                        f"/Ac/Consumption/L{i}/Power",
                    ),
                )
                self.p_cur_.append(
                    await self.intf.importer(
                        "com.victronenergy.system",
                        f"/Ac/ActiveIn/L{i}/Power",
                    ),
                )
            self.load = [0] * self.n_phase

            await self._update_vars()
            yield self

    @property
    def u_dc(self):
        "calculate u_dc but with internal resistance added"
        return self._u_dc.value + self.i_batt * self.r_int

    @property
    def batt_soc(self):
        "battery SoC as [0…1] (Dbus has 0…100)"
        return self._batt_soc.value / 100

    @property
    def i_batt(self):
        "battery current, inverted from Dbus"
        return -self._i_batt.value

    @property
    def i_pv(self):
        "PV current from Dbus"
        return self._i_pv.value

    @property
    def i_inv(self):
        "Inverter current from Dbus"
        return self._i_inv.value

    @property
    def p_inv(self):
        "AC power from inverter, calculated."
        return -self.p_cons - self.p_grid

    @property
    def p_cons(self):
        "Power from other AC consumers, between this inverter and the home meter."
        return -sum(x.value for x in self.p_cons_)

    @property
    def p_grid(self):
        "Power as measured by the grid meter."
        return sum(x.value for x in self.p_grid_)

    @property
    def ib_max(self):
        "Max battery current, discharging."
        # Remember that currents are measured from the PoV of the bus bar.
        # Thus this is the discharge current, current goes from the battery to the bus.
        if not self._ok_dis:
            return 0
        return self._ib_dis.value

    @property
    def ib_min(self):
        """
        Max battery current, charging. The charging current is negative
        so this is the lower limit.
        """
        # Remember that currents are measured from the PoV of the bus bar.
        # Thus the max charge current is negative, current goes into the battery.
        if not self._ok_chg:
            return 0
        return -self._ib_chg.value

    async def _update_vars(self):
        "Register Dbus readers"
        if not self.acc_vebus.value:
            logger.warning("NO vebus")
            return
        self.u_min = await self.intf.importer(self.s_battery.value, "/Info/BatteryLowVoltage")
        self.u_max = await self.intf.importer(self.s_battery.value, "/Info/MaxChargeVoltage")
        self._ib_chg = await self.intf.importer(self.s_battery.value, "/Info/MaxChargeCurrent")
        self._ib_dis = await self.intf.importer(self.s_battery.value, "/Info/MaxDischargeCurrent")
        self._ok_chg = await self.intf.importer(self.s_battery.value, "/Io/AllowToCharge")
        self._ok_dis = await self.intf.importer(self.s_battery.value, "/Io/AllowToDischarge")

        self.b_cap = (
            await self.intf.importer(self.s_battery.value, "/Capacity", createsignal=False)
        ).value
        self.p_set_ = []
        self.p_run_ = []
        for i in range(1, self.n_phase + 1):
            self.p_set_.append(
                await self.intf.importer(self.acc_vebus.value, f"/Hub4/L{i}/AcPowerSetpoint"),
            )
            self.p_run_.append(
                await self.intf.importer(self.acc_vebus.value, f"/Ac/ActiveIn/L{i}/P"),
            )
        self._p_inv = await self.intf.importer(
            self.acc_vebus.value,
            "/Ac/ActiveIn/P",
            eventCallback=self._trigger_step,
        )

    def _trigger_step(self, _sender, _path, _values):
        "signal that power adjustment step is complete"
        self._trigger.set()
        self._trigger = anyio.Event()

    async def trigger(self, sleep=3):
        "Wait for power adjustment step to complete"
        await anyio.sleep(sleep)
        with anyio.move_on_after(5):
            await self._trigger.wait()

    i_batt_avg = None

    async def _i_batt_task(self, evt):
        """
        Calculate a running average of the last four battery current values.
        """
        b_last = [None, None, None, None]
        while True:
            b_last.pop(0)
            b_last.append(self.i_batt)
            try:
                self.i_batt_avg = sum(b_last) / len(b_last)
            except (TypeError, ValueError):
                self.i_batt_avg = 0
            else:
                if evt is not None:
                    evt.set()
                    evt = None
            await anyio.sleep(1.1)

    i_pv_max = 0

    async def _avg_task(self):
        """
        Calculate exponential averages and decaying maxima.
        """
        # well that's a lie, currently we only track i_pv_max.
        while True:
            i = self.i_pv
            if i is None:
                continue
            if self.i_pv_max < self.i_pv:
                self.i_pv_max = self.i_pv
            elif self.i_pv_max > 1000 and self.i_pv < self.i_pv_max * self.pv_margin:
                # Owch, that was too fast
                pvm = self.i_pv / self.i_pv_max
                logger.error(
                    "PV went down too fast: margin factor set from %.2f to %.2f",
                    self.pv_margin,
                    pvm,
                )
                self.pv_margin = pvm
            else:
                self.i_pv_max += (self.i_pv - self.i_pv_max) / 20
            await anyio.sleep(0.9)

    async def _init_srv(self):
        """setup exports"""
        srv = self.srv
        evt = anyio.Event()
        self._tg.start_soon(self._i_batt_task, evt)
        self._tg.start_soon(self._avg_task)
        self._tg.start_soon(self._distkv_main)

        await srv.add_mandatory_paths(
            processname=__file__,
            processversion="0.1",
            connection="MoaT Inv " + self.acc_vebus.value.rsplit(".", 1)[1],
            deviceinstance="1",
            serial="123457",
            productid=123211,
            productname="MoaT Inverter Controller",
            firmwareversion="0.1",
            hardwareversion=None,
            connected=1,
        )

        # self.mode = await srv.add_path("/Mode", 0,
        #    description="Controller mode", writeable=True,
        #    onchangecallback=self._change_mode, gettextcallback=self._mode_name)

        # await self.mode.set_value(self._mode)
        self._mode_task = None
        self._mode_task_stopped = anyio.Event()
        self.dest_p = self.last_p = -self.p_grid - self.p_cons
        await evt.wait()

    #   async def _change_mode(self, path, value):
    #       await self.change_mode(value)

    async def change_mode(self, mode: str, data={}):  # noqa:B006
        "Switch mode to @mode, set config to @data"
        if self._mode is None or self._mode != mode:
            if self._change_mode_evt is None:
                raise DBusError("org.m_o_a_t.inv.too_early", "try again later")
            if mode not in self.MODES:
                raise DBusError("org.m_o_a_t.inv.unknown", "unknown mode")
            self._mode = mode
            # TODO verify the new mode's parameters
            if self._mode_task is not None:
                self._mode_task.cancel()
                await self._mode_task_stopped.wait()
            self._change_mode_evt.set()

        # TODO verify parameters for current mode
        self.op.update(data)
        for k, v in self.cfg["modes"].get(self._mode, {}).items():
            self.op.setdefault(k, v)
        return True

    async def change_mode_param(self, param: str, value: Any) -> bool:
        "Update a mode parameter"
        # pylint:disable=protected-access
        if not param or param[0] == "_" or param not in self.MODES[self._mode]._doc:  # noqa:SLF001
            raise DBusError("org.m_o_a_t.inv.unknown", f"unknown parameter for {self._mode}")
        self.op[param] = value
        return True

    async def _run_mode_task(self):
        "do the mode's work"
        # pylint:disable=protected-access
        if self._mode_task is not None:
            raise RuntimeError("cannot run two tasks")
        try:
            with anyio.CancelScope() as self._mode_task:
                m = self.MODES[self._mode]
                self.clear_state()
                self.set_state("mode", [m._name, self.op])  # noqa:SLF001
                self.op.update(self.cfg["modes"].get(self._mode, {}))
                await m(self).run()
        finally:
            logger.debug("MODE STOP %s", m._name)  # noqa:SLF001
            self._mode_task = None
            self._mode_task_stopped.set()

    async def _start_mode_task(self):
        self._tg.start_soon(self._run_mode_task)

    async def _solar_log(self):
        async with DbusMonitor(self._bus, self.MON) as mon:
            power = 0
            dkv = await self.distkv
            if dkv:
                val = await dkv.get(self.distkv_prefix / "solar" / "energy")
                if val and "value" in val:
                    power = val.value
            t = anyio.current_time()
            t_sol = t + 5
            mt = (
                min(time_until((n, "min")) for n in range(0, 60, 15))
                - datetime.now(tz=datetime.UTC)
            ).seconds
            print(mt)
            while True:
                n = 0
                while n < mt:  # 15min
                    t += 1
                    n += 1
                    cur_p = 0
                    for chg in mon.get_service_list("com.victronenergy.solarcharger"):
                        cur_p += mon.get_value(chg, "/Yield/Power") or 0
                    power += cur_p
                    self.solar_p = cur_p
                    if t >= t_sol:
                        if dkv:
                            for chg in mon.get_service_list("com.victronenergy.solarcharger"):
                                name = mon.get_value(chg, "/CustomName")
                                ni = name.find(" : ")
                                if ni >= 0:
                                    name = name[ni + 3 :]
                                pp = mon.get_value(chg, "/Yield/Power") or 0
                                await dkv.set(
                                    self.distkv_prefix / "solar" / "p" / name,
                                    pp,
                                    idem=True,
                                )
                            await dkv.set(self.distkv_prefix / "solar" / "p", cur_p, idem=True)
                            await dkv.set(
                                self.distkv_prefix / "solar" / "batt_pct",
                                self.batt_soc,
                                idem=True,
                            )
                            await dkv.set(
                                self.distkv_prefix / "solar" / "grid",
                                self.p_grid,
                                idem=True,
                            )
                        t_sol = t + 10
                    await anyio.sleep_until(t)
                print(power)
                if dkv:
                    await dkv.set(self.distkv_prefix / "solar" / "energy", power, idem=True)

                mt = 900

    async def _distkv_main(self):
        if "distkv" not in self.cfg:
            self._dkv_evt.set()
            self._dkv_evt = None
            return

        # TODO use a service scope instead
        # pylint: disable=import-outside-toplevel
        from moat.kv.client import open_client as distkv_client  # noqa: PLC0415

        try:
            async with distkv_client(**self.cfg["distkv"]) as dkv:
                self._dkv = dkv
                self.distkv_prefix = self.cfg["distkv"]["root"]
                self._dkv_evt.set()
                await anyio.sleep_forever()
        finally:
            self._dkv = None

    @property
    async def distkv(self):
        "fetch distkv object"
        if self._dkv_evt is None:
            return False
        await self._dkv_evt.wait()
        return self._dkv

    async def _init_intf(self):
        # setup BMS interface
        proxy = await self._bus.get_proxy_object(self.s_battery.value, "/bms")
        self._bms_intf = await proxy.get_interface("org.m_o_a_t.bms")

        self._batt_intf = []
        for i in range(await self._bms_intf.call_get_n_batteries()):
            proxy = await self._bus.get_proxy_object(self.s_battery.value, f"/bms/{i}")
            self._batt_intf.append(await proxy.get_interface("org.m_o_a_t.bms"))
        # TODO multiple batteries

    async def get_bms_work(self, poll: bool = False, clear: bool = False):
        "BMS: get current work"
        return await self._bms_intf.call_get_work(poll, clear)

    async def get_bms_voltages(self):
        "BMS: get current cell voltages"
        return [unwrap_dbus_dict(x) for x in await self._bms_intf.call_get_voltages()]

    async def get_bms_currents(self):
        "BMS: get currents"
        return await self._bms_intf.call_get_currents()

    async def get_bms_config(self):
        "BMS: get config data"
        return unwrap_dbus_dict(await self._bms_intf.call_get_config())

    async def set_bms_capacity(self, n: int, cap: float, loss: float, top: bool = False):
        "BMS: update actual capacity"
        if n != 0:
            raise NotImplementedError(n)
        return await self._batt_intf[n].call_set_capacity(cap, loss, top)

    async def run(self, mode=None):
        "do the work"
        self._change_mode_evt = anyio.Event()
        self._change_mode_evt.set()
        self._mode = mode or self.cfg["modes"]["default"]
        self._mode_task = None

        name = "org.m-o-a-t.power.inverter"

        async with (
            InvInterface(self) as self._ctrl,
            self.intf.service(name) as self._srv,
            anyio.create_task_group() as self._tg,
        ):
            self._tg.start_soon(self._init_intf)
            if not self.acc_vebus.value:
                logger.warning("VEBUS not known")
                await self.acc_vebus.refresh()
                if not self.acc_vebus.value:
                    raise RuntimeError("VEBUS not known")
            await self._init_srv()
            if not self.op.get("fake", False):
                self._tg.start_soon(self._solar_log)
            async with DbusName(
                self.bus,
                f"org.m_o_a_t.inv.{self.cfg.get('name', 'fake' if self.op.get('fake', False) else 'main')}",  # noqa:E501
            ):
                while True:
                    await self._change_mode_evt.wait()
                    self._change_mode_evt = None
                    await self._start_mode_task()
                    await anyio.sleep(30)
                    self._change_mode_evt = anyio.Event()

    def i_from_p(self, p, rev=False):
        """
        Calculate how much DC current a given inverter output would generate.

        Set ``rev`` if you want to know the DC current you'd need for a given AC power.
        """
        res = -p / self.u_dc
        if rev == (res < 0):
            res /= self.inv_eff
        else:
            res *= self.inv_eff
        # logger.debug("I %.1f from P %.0f %s", res, p, "R" if rev else "")
        return res

    def p_from_i(self, i, rev=False):
        """
        Calculate how much AC power to set for a given inverter DC current would generate.

        Set ``rev`` if you want to know the AC power you'd need for a given DC current.
        """
        res = -i * self.u_dc
        if rev == (res > 0):
            res /= self.inv_eff
        else:
            res *= self.inv_eff
        # logger.debug("P %.0f from I %.1f %s", res, i, "R" if rev else "")
        return res

    def calc_batt_i(self, i):
        """
        Calculate how much AC power to set for a given battery current
        """
        logger.debug("Want bat I: %.1f", i)
        ii = max(self.ib_min, min(i, self.ib_max))
        if ii != i:
            logger.debug("Adj: bat I: %f", ii)
        # i_pv+i_batt+i_inv == zero
        return self.calc_inv_i(-ii - self.i_pv)

    def calc_inv_i(self, i):
        """
        Calculate how much AC power to set for a given inverter current
        """
        logger.debug("Want inv I: %.1f", i)
        return self.calc_inv_p(self.p_from_i(i))

    def calc_grid_p(self, p, excess=None):
        """
        Set power from/to the grid. Positive = take power.

        p_cons+p_grid+p_inv == zero by definition.
        """
        logger.debug("Want grid P: %.0f", p)

        p = -self.p_cons - p
        return self.calc_inv_p(p, excess=excess)

    def calc_inv_p(self, p, excess=None, phase=None):
        """
        Calculate intended inverter/charger power.

        This code adheres to quite a few constraints, thus might not end
        up doing what you want it to do.
        """
        lims = []
        no_lims = []
        p_info = dict(
            limits=lims,
            # non_limits=no_lims,
            init=p,
        )

        logger.debug("WANT inv P: %.0f", p)
        op = p

        i_inv = self.i_from_p(p, rev=True)
        i_batt = -i_inv - self.i_pv

        # if the PV input is close to the maximum, increase power
        i_max = self.ib_max - i_inv
        lim = dict(
            rule="I_PVD",
            pvmax=self.i_pv_max,
            pvdelta=self.pv_delta,
            imax=i_max,
            ib=i_batt,
            lim="pvmax>pvdelta, imax-ib<pvdelta",
        )
        if self.i_pv_max > self.pv_delta and i_max - i_batt < self.pv_delta:
            lim["fix"] = "ib=imax-pv_delta"
            lim["res"] = i_batt = i_max - self.pv_delta
            lims.append(lim)
        else:
            no_lims.append(lim)

        # if we're close to the max voltage, slow down / stop early
        i_maxchg = (
            self.b_cap
            / self.cap_scale
            * ((0 if self.top_off else self.umax_diff) - (self.u_max.value - self.u_dc))
            / self.umax_diff
        )
        lim = dict(
            rule="U_MAX",
            max=i_maxchg,
            cap_lim=self.b_cap / self.cap_scale,
            range=(
                0 if self.top_off else self.umax_diff,
                self.u_max.value - self.u_dc,
                self.umax_diff,
            ),
            umax=self.u_max.value,
            udc=self.u_dc,
            ib=i_batt,
            lim="ib<max",
        )
        if i_batt < i_maxchg:
            lim["fix"] = "ib=max"
            i_batt = i_maxchg
            i_inv = -i_batt - self.i_pv
            lims.append(lim)
            lim["res"] = {"batt": i_batt, "inv": i_inv}
        else:
            no_lims.append(lim)

        # On the other side, if we're close to the min voltage, limit discharge rate.
        i_maxdis = (
            -self.b_cap
            / self.cap_scale
            * (self.umin_diff - (self.u_dc - self.u_min.value))
            / self.umin_diff
        )
        lim = dict(
            rule="U_MIN",
            min=i_maxdis,
            cap_lim=self.b_cap / self.cap_scale,
            range=(self.umin_diff, self.u_dc - self.u_min.value),
            umin=self.u_min.value,
            udc=self.u_dc,
            ib=i_batt,
            lim="ib<min",
        )
        if i_batt > i_maxdis:
            lim["fix"] = "ib=min"
            i_batt = i_maxdis
            i_inv = -i_batt - self.i_pv
            lim["res"] = {"batt": i_batt, "inv": i_inv}
            lims.append(lim)
        else:
            no_lims.append(lim)

        # The system tells the solar chargers how much they may deliver.
        # However, we want to leave a margin for them
        # so that we can actually notice when PV output increases.
        #
        #             'limits': [{'d': 17.970052750227836,
        #                         'fix': 'ib-=d',
        #                         'ibmin': -83.2819387025245,
        #                         'inv': -56.74800979149463,
        #                         'ipv': 128.00000124424696,
        #                         'lim': 'max-ipv<pvdelta',
        #                         'max': 140.02994849401912,
        #                         'pvdelta': 30,
        #                         'res': {'batt': -89.22204420298016,
        #                                 'inv': -38.7779570412668},
        #                         'rule': 'I_MAX'}],

        i_pv_max = -self.ib_min - i_inv  # this is what Venus systemcalc sets the PV max to
        lim = dict(
            rule="I_MAX",
            max=i_pv_max,
            ipv=self.i_pv,
            pvdelta=self.pv_delta,
            ibmin=self.ib_min,
            inv=i_inv,
            lim="max-ipv<pvdelta",
        )
        if i_pv_max - self.i_pv < self.pv_delta:
            lim["d"] = d = self.pv_delta - (i_pv_max - self.i_pv)
            lim["fix"] = "ib-=d"
            i_batt -= d
            i_inv = -i_batt - self.i_pv
            lim["res"] = {"batt": i_batt, "inv": i_inv}
            lims.append(lim)
        else:
            no_lims.append(lim)

        # Now check some AC limits.
        p = self.p_from_i(i_inv)

        # Don't overload the charger.
        lim = dict(
            rule="P_MIN",
            p=p,
            min=self.pg_min,
            lim="p<min",
        )
        if p < self.pg_min:
            lim["fix"] = "p=min"
            lim["res"] = p = self.pg_min
            lims.append(lim)
        else:
            no_lims.append(lim)

        # Don't overload the inverter.
        lim = dict(
            rule="P_MAX",
            p=p,
            max=self.pg_max,
            lim="p>max",
        )
        if p > self.pg_max:
            lim["fix"] = "p=max"
            lim["res"] = p = self.pg_max
            lims.append(lim)
        else:
            no_lims.append(lim)

        # We want to be on the safe side with varying PV input. Consider this state:
        # * PV delivers 60 A
        # * battery discharge maximum is 60 A
        # * thus we think it's safe for the inverter to take 120A
        # * now PV output drops to 20A due to an ugly black cloud
        # * 40A over the discharge limit is probably enough to trip the BMS
        # * … owch.
        #
        i_inv = self.i_from_p(p, rev=True)
        i_pv_min = self.i_pv_max * self.pv_margin
        lim = dict(
            rule="I_MIN",
            inv=i_inv,
            pvmin=i_pv_min,
            ibmax=self.ib_max,
            lim="-inv-pvmin>ibmax",
            max=self.i_pv_max,
            margin=self.pv_margin,
        )
        if -i_inv - i_pv_min > self.ib_max:
            i_inv = -i_pv_min - self.ib_max
            i_batt = -i_inv - self.i_pv
            lim["fix"] = "inv=-pvmin-ibmax"
            lim["res"] = {"batt": i_batt, "inv": i_inv}
            lims.append(lim)
        else:
            no_lims.append(lim)

        # Don't push more into the battery than allowed.
        lim = dict(
            rule="IB_ERR_L",
            batt=i_batt,
            min=self.ib_min,
            lim="batt<min",
        )
        if i_batt < self.ib_min:
            lim["fix"] = ("batt=min",)
            i_batt = self.ib_min
            i_inv = -i_batt - self.i_pv
            lim["res"] = {"batt": i_batt, "inv": i_inv}
        # no add to no_lims because it's too obvious

        # Don't pull more from the battery than allowed.
        lim = dict(
            rule="IB_ERR_H",
            batt=i_batt,
            max=self.ib_max,
            lim="batt>max",
        )
        if i_batt > self.ib_max:
            lim["fix"] = ("batt=max",)
            i_batt = self.ib_max
            i_inv = -i_batt - self.i_pv
            lim["res"] = {"batt": i_batt, "inv": i_inv}
        # no add to no_lims because it's too obvious

        # back to the AC side
        p = self.p_from_i(i_inv)

        # Don't exceed what we're allowed to feed to the grid.
        if excess is None:
            no_lims.append({"rule": "P_EXC", "exc": "-"})
        else:
            lim = dict(
                rule="P_EXC",
                lim="p>op+exc, p>0",
                p=p,
                op=op,
                exc=excess,
            )
            if p > 0 and p > op + excess:
                lim["fix"] = "p=op+exc"
                lim["res"] = p = op + excess
                lims.append(lim)
            else:
                no_lims.append(lim)

        # This part ensures that we don't take too-huge steps towards
        # the new value, which would cause instabilities.
        if self.f_delta <= self.batt_soc <= 1 - self.f_delta or self.small_p_step(self.last_p, p):
            # Small(ish) change from last target, or far enough away from SoC bounds not to care.
            # Implement directly.
            np = p
            self.step = 1
        else:
            if self.small_p_step(self.dest_p, p):
                # the goal is roughly the same as last time
                self.step += 1
            else:
                # reset the step counter if the goal has changed significantly
                self.step = 2

            # The first step goes 1/3rd towards the destination (if f_step
            # is 1/3). The second step, ~ halfway. This tries to strike a
            # hopefully-reasonable balance between slow exponential decay
            # and going too fast.
            pd = (p - self.last_p) * self.f_step ** (2 / self.step)

            # If the scaled-off step is < p_step, use that instead:
            # smaller steps should be taken last.
            if abs(pd) < self.p_step:
                pd = self.p_step * (-1 if pd < 0 else 1)
            np = self.last_p + pd
            lim["step"] = (pd, self.last_p, p, np)
            logger.debug("P_GRAD: %.0f > %.0f = %.0f", self.last_p, p, np)

        p_info["dest"] = self.dest_p = p
        self.last_p = np

        p_info["setpoint"] = np

        if phase is None and self.n_phase > 1:
            ps = self.to_phases(np)
        else:
            ps = [0] * self.n_phase
            ps[phase - 1 if phase else 0] = np
        p_info["inv_phases"] = ps
        self.set_state("inverter", p_info)
        return ps

    def small_p_step(self, p, q) -> bool:
        """
        True if th delta between @p and @q is to be considered small
        """
        if abs(p - q) < self.p_step:
            return True
        if (p > 0) != (q > 0):
            return False
        if 10 / 12 < ((self.p_step + abs(p)) / (self.p_step + abs(q))) < 12 / 10:
            return True
        return False

    def to_phases(self, p):
        """
        Distribute a given power to the phases so that the result is balanced,
        or at least as balanced as possible given that to feed in from one phase
        while sending energy out to another phase is a waste of energy.

        This step does not change the total, assuming the per-phase limits
        are not exceeded.
        """

        self.load = [-b.value for b in self.p_cons_]
        load_avg = sum(self.load) / self.n_phase

        ps = [p / self.n_phase - (g - load_avg) for g in self.load]
        ps = balance(ps, val_min=-self.p_per_phase, val_max=self.p_per_phase)
        return ps

    async def set_inv_ps(self, ps):
        "set inverter power from array"
        # OK, we're safe, implement
        if self.op.get("fake", False):
            if self.n_phase > 1:
                logger.error(
                    "NO-OP SET inverter %.0f ∑ %s",
                    -sum(ps),
                    " ".join(f"{-x:.0f}" for x in ps),
                )
            else:
                logger.error("NO-OP SET inverter %.0f", -ps[0])
            return

        if self.n_phase > 1:
            logger.info("SET inverter %.0f ∑ %s", -sum(ps), " ".join(f"{-x:.0f}" for x in ps))
        else:
            logger.info("SET inverter %.0f", -ps[0])

        for p, v in zip(self.p_set_, ps, strict=True):
            await p.set_value(-v)
            # Victron Multiplus: negative=inverting: positive=charging
            # This code: negative=takes from AC, positive=feeds to AC power
            # thus this is inverted

    @property
    def srv(self):
        "interface service accessor"
        return self._srv

    @property
    def intf(self):
        "interface accessor"
        return self._intf

    @property
    def tg(self):
        "taskgroup accessor"
        return self._tg


class InvModeBase:
    "Base class for inverter control modules"

    def __init__(self, intf):
        self.intf = intf
        self.ps = None
        self.ps_min = [-999999999] * intf.n_phase
        self.ps_max = [999999999] * intf.n_phase
        self.running = False

    # p_set_
    #   the power we want this Multi to get/emit. Negative.
    # p_cur_
    #   The power actually going to/from the grid. Positive.
    # p_run_
    #   The power the Multi is feeding us. Negative.
    # p_cons_
    #   The power other consumers are taking from the bus. Negative.
    async def set_inv_ps(self, ps):
        """
        Apply per-inverter power settings.

        This method observes fixed maximum power values as well as
        actual power output caps. It redistributes intended power
        so that the grid feed is balanced, secondary to achieving
        the intended inverter load.

        """
        intf = self.intf

        # This convoluted step thinks about inverter output limits.
        # Specifically, if one of them is overloaded and another is not
        # we distribute the excess to others inverters.
        #
        # Yes this may result in one grid phase going in while others
        # go out, but that could already happen anyway and should not
        # materially increase grid load differences across phases.
        if self.running and intf.n_phase > 1:
            pd_min = pd_max = 0
            d_min = d_max = 0
            ops = ps

            # First pass: determine the invertes' current operational limits.
            for i in range(intf.n_phase):
                p = ps[i]
                p_set = intf.p_set_[i].value
                # p_cur = intf.p_cur_[i].value
                p_run = intf.p_run_[i].value
                # p_cons = -intf.p_cons_[i].value
                p_min = self.ps_min[i]
                p_max = self.ps_max[i]
                # logger.debug("%.0f %.0f %.0f %.0f %.0f %.0f", p_set,p_cur,p_run,p_cons,p_min,p_max)  # noqa:E501

                if p_set < 0:
                    if p_set < p_run - 20:
                        # power out seems to be limited
                        self.ps_min[i] = p_min = p_run
                        if p < p_min:
                            pd_min += p_min - p - 50
                    elif p_min == -999999999:
                        pass
                    elif p_min >= p_run - 10:
                        self.ps_min[i] = -999999999
                        # unknown, we don't ask for it
                    elif p < 0 and p < p_min - 50:
                        # this will go below the limit
                        pd_min += p_min - p - 50

                elif p_set > 0:  # same in reverse for p_max
                    if p_set > p_run + 20:
                        self.ps_max[i] = p_max = p_run
                        if p > p_max:
                            pd_max += p - p_max + 50
                    elif p_max == 999999999:
                        pass
                    elif p_max <= p_run - 10:
                        self.ps_max[i] = 999999999
                    elif p > 0 and p > p_max - 50:
                        pd_max += p_max - p + 50

            # Second pass. Distribute difference to lower-powered phases.
            pa = list(enumerate(ps))
            if pd_min > 0:
                # We sort by "worst-hit device last". Going through the
                # list from the end, the delta is accumulated in d_min and
                # then distributed to the remaining devices, if any.
                #
                # We add a fudge of 50W so that we can discover (in the next
                # round's first pass) whether the limit has been lifted.
                # 50W is too high for small batteries, but if you really
                # do run a multiphase system on a 12V 20A battery, you're
                # going to have worse problems than this. :-P
                pa.sort(key=lambda x: -ps[x[0]] + self.ps_min[x[0]])
                # logger.debug("MIN Pre %s", pa)
                pb = []
                d_min = 0
                while pa:
                    i, v = pa.pop()
                    p_run = intf.p_run_[i].value
                    p_min = self.ps_min[i]
                    if v < p_min:  # over the limit
                        d_min += p_min - v
                        v = p_min - 50
                    else:
                        pp = d_min / (len(pa) + 1)
                        if v - pp < p_min:  # limited by min
                            d_min -= v - p_min
                            v = p_min - 50
                        else:  # can use pp
                            d_min -= pp + 50
                            v -= pp + 50
                    pb.append((i, v))
                pa = pb

            if pd_max > 0:
                # same as above for taking power from the grid.
                # Not yet tested because it's summer.
                pa.sort(key=lambda x: ps[x[0]] - self.ps_max[x[0]])
                # logger.debug("MAX Pre %s", pa)
                pb = []
                d_max = 0
                while pa:
                    i, v = pa.pop()
                    p_run = intf.p_run_[i].value
                    p_max = self.ps_max[i]
                    if v > p_max:  # over the limit
                        v = p_max + 50
                    else:
                        pp = d_max / (len(pa) + 1)
                        if v + pp > p_max:  # limited by max
                            d_max -= p_max - v
                            v = p_max + 50
                        else:  # can use pp
                            d_max -= pp + 50
                            v += pp + 50
                    pb.append((i, v))
                pa = pb
            ps = [x[1] for x in sorted(pa, key=lambda x: x[0])]

            if ops != ps and (d_min > 0 or d_max > 0):
                # logger.debug("START %s",ops)
                if d_min > 0:
                    logger.debug("PD_MIN: P %.0f, want %.0f", sum(x[1] for x in pa), d_min)
                if d_max > 0:
                    logger.debug("PD_MAX: P %.0f, want %.0f", sum(x[1] for x in pa), d_max)
                # togger.debug("END %s",ps)

        await intf.set_inv_ps(ps)

        p = intf.p_inv
        await intf.trigger()
        n = 0
        nt = False
        while n < 10:
            await intf.trigger()
            pp = intf.p_inv
            logger.debug("now %.0f", pp)
            if abs(pp - p) < intf.p_step:
                if nt:
                    break
                nt = True
            else:
                nt = False
            p = pp
            n += 1
        self.running = True


# from ._utils import _loader
def _loader(path, cls, reg):
    "Load submodule"
    # pylint: disable=import-outside-toplevel
    from importlib import import_module  # noqa: PLC0415
    from pathlib import Path  # noqa: PLC0415

    def _imp(name):
        m = import_module("." + name, package=_loader.__module__)  # noqa:F821
        for n in m.__all__:
            c = getattr(m, n)
            if isinstance(c, type) and issubclass(c, cls) and hasattr(c, "_name"):
                reg(c)

    path = Path(__path__[0])
    for filename in path.glob("*"):
        if filename.name[0] in "._":
            continue
        if filename.name.endswith(".py"):
            _imp(filename.name[:-3])
        elif (path / filename / "__init__.py").is_file():
            _imp(filename.name)


_loader(__path__[0], InvModeBase, InvControl.register)
del _loader



#! /usr/bin/python3 -u

from argparse import ArgumentParser
import dbus
import dbus.mainloop.glib
import faulthandler
import os
import pymodbus.constants
from settingsdevice import SettingsDevice
import signal
import time
import traceback
from vedbus import VeDbusService
from gi.repository import GLib

import device
import mdns
import probe
from scan import *
from utils import *

# import carlo_gavazzi
# import ev_charger
# import smappee
import TWE_Eastron_SDM120

import logging

log = logging.getLogger()

NAME = os.path.basename(__file__)
VERSION = "1.13"

__all__ = ["NAME", "VERSION"]

pymodbus.constants.Defaults.Timeout = 0.5

MODBUS_TCP_PORT = 502
MODBUS_TCP_UNIT = 1

MAX_ERRORS = 5
FAILED_INTERVAL = 10
MDNS_CHECK_INTERVAL = 5
MDNS_QUERY_INTERVAL = 60
SCAN_INTERVAL = 600
UPDATE_INTERVAL = 250

if_blacklist = [
    "ap0",
]


def percent(path, val):
    return "%d%%" % val


class Client(object):
    def __init__(self, name):
        self.name = name
        self.devices = []
        self.failed = []
        self.failed_time = 0
        self.scanner = None
        self.scan_time = time.time()
        self.auto_scan = False
        self.err_exit = False
        self.keep_failed = True
        self.svc = None

    def start_scan(self, full=False):
        if self.scanner:
            return

        log.info("Starting background scan")

        s = self.new_scanner(full)

        if s.start():
            self.scanner = s

    def stop_scan(self):
        if self.scanner:
            self.scanner.stop()

    def scan_update(self):
        devices = self.scanner.get_devices()

        for d in devices:
            if d in self.devices:
                d.destroy()
                continue

            try:
                d.init(self.dbusconn)
                d.nosave = False
                self.devices.append(d)
            except Exception:
                log.info("Error initialising %s, skipping", d)
                traceback.print_exc()

        self.save_devices()

    def scan_complete(self):
        self.scan_time = time.time()

        if not self.devices and self.err_exit:
            os._exit(1)

    def set_scan(self, path, val):
        if val:
            self.start_scan()
        else:
            self.stop_scan()

        return True

    def update_device(self, dev):
        try:
            dev.update()
            dev.err_count = 0
        except Exception:
            dev.err_count += 1
            if dev.err_count == MAX_ERRORS:
                log.info("Device %s failed", dev)
                if self.err_exit:
                    os._exit(1)
                self.devices.remove(dev)
                if not dev.nosave:
                    self.failed.append(str(dev))
                dev.destroy()

    def probe_devices(self, devlist, nosave=False):
        devs = set(devlist) - set(self.devices)
        devs, failed = probe.probe(devs)

        for d in devs:
            try:
                d.init(self.dbusconn)
                d.nosave = nosave
                self.devices.append(d)
            except Exception:
                failed.append(str(d))
                d.destroy()

        return failed

    def save_devices(self):
        devs = filter(lambda d: not d.nosave, self.devices)
        devstr = ",".join(sorted(list(map(str, devs)) + self.failed))
        if devstr != self.settings["devices"]:
            self.settings["devices"] = devstr

    def update_devlist(self, old, new):
        old = set(old.split(","))
        new = set(new.split(","))
        cur = set(self.devices)
        rem = old - new

        for d in rem & cur:
            dd = self.devices.pop(self.devices.index(d))
            dd.destroy()

        self.failed = self.probe_devices(new)
        self.save_devices()

    def setting_changed(self, name, old, new):
        if name == "devices":
            self.update_devlist(old, new)
            return

    def init(self, force_scan):
        settings_path = "/Settings/ModbusClient/" + self.name
        SETTINGS = {
            "devices": [settings_path + "/Devices", "", 0, 0],
            "autoscan": [settings_path + "/AutoScan", self.auto_scan, 0, 1],
        }

        self.dbusconn = private_bus()

        log.info("Waiting for localsettings")
        self.settings = SettingsDevice(self.dbusconn, SETTINGS, self.setting_changed, timeout=10)

        self.update_devlist("", self.settings["devices"])

        if not self.keep_failed:
            self.failed = []

        scan = force_scan

        if not self.devices or self.failed:
            if self.settings["autoscan"]:
                scan = True

        if scan:
            self.start_scan(force_scan)

    def update(self):
        if self.scanner:
            if self.svc:
                self.svc["/Scan"] = self.scanner.running
                self.svc["/ScanProgress"] = 100 * self.scanner.done / self.scanner.total

            self.scan_update()

            if not self.scanner.running:
                self.scan_complete()
                self.scanner = None
                if self.svc:
                    self.svc["/ScanProgress"] = None

        for d in self.devices:
            self.update_device(d)

        if self.failed:
            now = time.time()

            if now - self.failed_time > FAILED_INTERVAL:
                self.failed = self.probe_devices(self.failed)
                self.failed_time = now

            if self.settings["autoscan"]:
                if now - self.scan_time > SCAN_INTERVAL:
                    self.start_scan()

    def update_timer(self):
        try:
            self.update()
        except Exception:
            log.error("Uncaught exception in update")
            traceback.print_exc()

        return True


class NetClient(Client):
    def __init__(self, proto):
        Client.__init__(self, proto)
        self.proto = proto

    def new_scanner(self, full):
        return NetScanner(self.proto, MODBUS_TCP_PORT, MODBUS_TCP_UNIT, if_blacklist)

    def init(self, *args):
        super(NetClient, self).init(*args)

        svcname = "com.victronenergy.modbusclient.%s" % self.name
        self.svc = VeDbusService(svcname, self.dbusconn)
        self.svc.add_path("/Scan", False, writeable=True, onchangecallback=self.set_scan)
        self.svc.add_path("/ScanProgress", None, gettextcallback=percent)

        self.mdns = mdns.MDNS()
        self.mdns.start()
        self.mdns_check_time = 0
        self.mdns_query_time = 0

    def update(self):
        super(NetClient, self).update()

        now = time.time()

        if now - self.mdns_query_time > MDNS_QUERY_INTERVAL:
            self.mdns_query_time = now
            self.mdns.req()

        if now - self.mdns_check_time > MDNS_CHECK_INTERVAL:
            self.mdns_check_time = now
            maddr = self.mdns.get_devices()
            if maddr:
                units = probe.get_units("tcp")
                d = []
                for a in maddr:
                    d += ["tcp:%s:%s:%d" % (a[0], a[1], u) for u in units]
                self.probe_devices(d, nosave=True)

        return True


class SerialClient(Client):
    def __init__(self, tty, rate, mode):
        Client.__init__(self, tty)
        self.tty = tty
        self.rate = rate
        self.mode = mode
        self.auto_scan = True
        self.keep_failed = False

    def new_scanner(self, full):
        return SerialScanner(self.tty, self.rate, self.mode, full=full)


def main():
    parser = ArgumentParser(add_help=True)
    parser.add_argument("-d", "--debug", help="enable debug logging", action="store_true")
    parser.add_argument("-f", "--force-scan", action="store_true")
    parser.add_argument("-m", "--mode", choices=["ascii", "rtu"], default="rtu")
    parser.add_argument("-r", "--rate", type=int, action="append")
    parser.add_argument("-s", "--serial")
    parser.add_argument("-x", "--exit", action="store_true", help="exit on error")

    args = parser.parse_args()

    logging.basicConfig(
        format="%(levelname)-8s %(message)s",
        level=(logging.DEBUG if args.debug else logging.INFO),
    )

    logging.getLogger("pymodbus.client.sync").setLevel(logging.CRITICAL)

    signal.signal(signal.SIGINT, lambda s, f: os._exit(1))
    faulthandler.register(signal.SIGUSR1)

    dbus.mainloop.glib.threads_init()
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    mainloop = GLib.MainLoop()

    if args.serial:
        tty = os.path.basename(args.serial)
        client = SerialClient(tty, args.rate, args.mode)
    else:
        client = NetClient("tcp")

    client.err_exit = args.exit
    client.init(args.force_scan)

    GLib.timeout_add(UPDATE_INTERVAL, client.update_timer)
    mainloop.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from moat.ems.victron.dbus import Dbus
import anyio


def mon(*a):
    print(a)


N = "test.victron.sender"
V = "/Some/Value"


async def main():
    async with Dbus() as bus:
        v = await bus.importer(N, V, eventCallback=mon)

        print("Initial value:", v.value)
        await anyio.sleep(999)


anyio.run(main, backend="trio")



# VenusOS module for support of Carlo Gavazzi EM24 PFB- and X-Models with Modbus-RTU-support
# - Meter has be be connected by a ModbusTCP gateway and cannot be connected directly by serial interface to VenusGX
# - Library might work also with PFA devices, but generate an error for application setting.
#
# Community contribution by Thomas Weichenberger
# Version 1.0 - 2022-01-02
#
# Thanks to Victron for their open platform and especially for the support of Matthijs Vader
# For any usage a donation to seashepherd.org with an amount of 5 USD/EUR/GBP or equivalent is expected

import logging
import EM24RTU_device as device
import probe
from register import *

log = logging.getLogger()

nr_phases = [3, 3, 2, 1, 3]

phase_configs = [
    "3P.n",
    "3P.1",
    "2P",
    "1P",
    "3P",
]

hardware_version = [
    "EM24DINAV93XO2PFA",
    "unknown",
    "EM24DINAV93XISPFA",
    "EM24DINAV93XXXPFA",
    "EM24DINAV23XO2PFA",
    "EM24DINAV23XISPFA",
    "EM24DINAV23XXXPFA",
    "EM24DINAV53xO2PFA",
    "unknown",
    "EM24DINAV53xISPFA",
    "EM24DINAV53xXXPFA",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "EM24DINAV93XISX",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "unknown",
    "EM24DINAV53DISX",
    "unknown",
    "unknown",
    "EM24DINAV63xO2X",
    "EM24DINAV63xR2X",
    "EM24DINAV63xISX",
    "EM24DINAV63xXXX",
]

switch_positions = [
    "kVARh",
    "2",
    "1",
    "Locked",
]


class EM24RTU_Meter(device.EnergyMeter):
    productid = 0xB017
    productname = "Carlo Gavazzi EM24RTU Energy Meter"
    min_timeout = 0.5

    def __init__(self, *args):
        super(EM24RTU_Meter, self).__init__(*args)

        self.info_regs = [
            Reg_u16(0x0302, "/HardwareVersion", text=hardware_version, write=(0, 78)),
            Reg_u16(0x0303, "/FirmwareVersion"),
            Reg_u16(0x1102, "/PhaseConfig", text=phase_configs, write=(0, 4)),
            Reg_text(0x1300, 7, "/Serial"),
        ]

    def phase_regs(self, n):
        s = 2 * (n - 1)
        return [
            Reg_s32l(0x0000 + s, "/Ac/L%d/Voltage" % n, 10, "%.1f V"),
            Reg_s32l(0x000C + s, "/Ac/L%d/Current" % n, 1000, "%.1f A"),
            Reg_s32l(0x0012 + s, "/Ac/L%d/Power" % n, 10, "%.1f W"),
            Reg_s32l(0x0046 + s, "/Ac/L%d/Energy/Forward" % n, 10, "%.1f kWh"),
        ]

    def device_init(self):
        # make sure application is set to H
        appreg = Reg_u16(0x1101)
        if self.read_register(appreg) != 7:
            self.write_register(appreg, 7)

            # read back the value in case the setting is not accepted
            # for some reason
            if self.read_register(appreg) != 7:
                log.error("%s: failed to set application to H", self)
                return

        self.read_info()

        phases = nr_phases[int(self.info["/PhaseConfig"])]

        regs = [
            Reg_s32l(0x0028, "/Ac/Power", 10, "%.1f W"),
            Reg_u16(0x0037, "/Ac/Frequency", 10, "%.1f Hz"),
            Reg_s32l(0x003E, "/Ac/Energy/Forward", 10, "%.1f kWh"),
            Reg_s32l(0x005C, "/Ac/Energy/Reverse", 10, "%.1f kWh"),
        ]

        if phases == 3:
            regs += [
                Reg_mapu16(0x0036, "/PhaseSequence", {0: 0, 0xFFFF: 1}),
            ]

        for n in range(1, phases + 1):
            regs += self.phase_regs(n)

        self.data_regs = regs

    def dbus_write_register(self, reg, path, val):
        super(EM24RTU_Meter, self).dbus_write_register(reg, path, val)
        self.sched_reinit()

    def get_ident(self):
        return "cg_%s" % self.info["/Serial"]


models = {
    70: {
        "model": "EM24-DIN AV9",
        "handler": EM24RTU_Meter,
    },
    71: {
        "model": "EM24-DIN AV2",
        "handler": EM24RTU_Meter,
    },
    72: {
        "model": "EM24-DIN AV5",
        "handler": EM24RTU_Meter,
    },
    73: {
        "model": "EM24-DIN AV6",
        "handler": EM24RTU_Meter,
    },
}

probe.add_handler(probe.ModelRegister(Reg_u16(0x000B), models, methods=["tcp", "rtu"], units=[1]))



"utility stuff"

from __future__ import annotations


def balance(a, val_min=None, val_max=None):
    """
    Takes an array. If there are positive *and* negative values in it, add the
    negative to the positive values so that the deltas between values are constant.

    Then, if some values exceed the val_min or val_max, cap it and distribute the delta.

    See `test/test_balance` for examples.
    """
    if not a:
        return a

    sl = sum(-x for x in a if x < 0)
    sh = sum(x for x in a if x > 0)

    rev = sl > sh
    if rev:
        a = [-x for x in a]
        d = sh
    else:
        d = sl

    # sort
    a = sorted(enumerate(a), key=lambda x: -x[1])

    # drop the negatives, if there are any
    if a[-1][1] >= 0:
        ra = a
        ra.reverse()
    else:
        ra = []
        while a:
            i, v = a.pop()
            if v <= 0:
                ra.append((i, 0))
                continue
            rd = d / (len(a) + 1)
            if rd >= v:
                d -= v
                ra.append((i, 0))
                continue
            ra.append((i, v - rd))
            d -= rd

    m = val_min if rev else val_max
    if m is None:
        a = ra
    else:
        if not isinstance(m, (tuple, list)):
            m = [m] * len(ra)
        a = []
        d = 0
        while ra:
            i, v = ra.pop()
            rd = d / (len(ra) + 1)
            mi = -m[i] if rev else m[i]
            if v + rd > mi:
                d += v - mi
                a.append((i, mi))
                continue
            a.append((i, v + rd))
            d -= rd

    a.sort(key=lambda x: x[0])
    if rev:
        return [-v for i, v in a]
    else:
        return [v for i, v in a]



# VenusOS module for support of ABB B21, B23 and B24 meters
#
# Community contribution by Thomas Weichenberger
# Version 1.0 - 2022-01-06
#
# Thanks to Victron for their open platform and especially for the support of Matthijs Vader
# For any usage a donation to seashepherd.org with an amount of 5 USD/EUR/GBP or equivalent is expected

import logging
import device
import probe
from register import *

log = logging.getLogger()


class Reg_u64b(Reg_num):
    coding = (">Q", ">4H")
    count = 4
    rtype = float


nr_phases = [1, 3, 3]

phase_configs = [
    "1P",
    "3P.n",
    "3P",
]


class ABB_B2x_Meter(device.EnergyMeter):
    productid = 0xB017
    productname = "ABB B2x-Series Power Meter"
    min_timeout = 0.5

    def __init__(self, *args):
        super(ABB_B2x_Meter, self).__init__(*args)

        self.info_regs = [
            Reg_text(0x8960, 6, "/HardwareVersion"),
            Reg_text(0x8908, 8, "/FirmwareVersion"),
            Reg_u32b(0x8900, "/Serial"),
        ]

    def phase_regs(self, n):
        s2 = 0x0002 * (n - 1)
        s4 = 0x0004 * (n - 1)

        return [
            Reg_u32b(0x5B00 + s2, "/Ac/L%d/Voltage" % n, 10, "%.1f V"),
            Reg_u32b(0x5B0C + s2, "/Ac/L%d/Current" % n, 100, "%.1f A"),
            Reg_s32b(0x5B16 + s2, "/Ac/L%d/Power" % n, 100, "%.1f W"),
            Reg_u64b(0x5460 + s4, "/Ac/L%d/Energy/Forward" % n, 100, "%.1f kWh"),
            Reg_u64b(0x546C + s4, "/Ac/L%d/Energy/Reverse" % n, 100, "%.1f kWh"),
        ]

    def device_init(self):
        self.read_info()

        phases = 3
        # phases = nr_phases[int(self.info['/PhaseConfig'])]

        regs = [
            Reg_s32b(0x5B14, "/Ac/Power", 100, "%.1f W"),
            Reg_u16(0x5B2C, "/Ac/Frequency", 100, "%.1f Hz"),
            Reg_u64b(0x5000, "/Ac/Energy/Forward", 100, "%.1f kWh"),
            Reg_u64b(0x5004, "/Ac/Energy/Reverse", 100, "%.1f kWh"),
        ]

        for n in range(1, phases + 1):
            regs += self.phase_regs(n)

        self.data_regs = regs

    def get_ident(self):
        return "cg_%s" % self.info["/Serial"]


models = {
    16946: {
        "model": "ABB_B2x",
        "handler": ABB_B2x_Meter,
    },
}


probe.add_handler(probe.ModelRegister(Reg_u16(0x8960), models, methods=["tcp", "rtu"], units=[1]))



#!/usr/bin/env python3

import sys
import anyio
from moat.ems.victron.dbus import Dbus


def mon(*a):
    print(a)


async def main():
    async with Dbus() as bus, bus.service("test.victron.sender") as srv:
        v = await srv.add_path("/Some/Value", None, description="Test", onchangecallback=mon)
        await srv.add_mandatory_paths(
            processname=__file__,
            processversion="0.1",
            connection="test",
            deviceinstance=0,
            productid=None,
            productname=None,
            firmwareversion=None,
            hardwareversion=None,
            connected=1,
        )

        print("Sending")
        n = 0
        while True:
            n += 1
            await anyio.sleep(n)
            await v.local_set_value(n)


try:
    anyio.run(main, backend="trio")
except KeyboardInterrupt:
    print("Interrupted.", file=sys.stderr)



from copy import copy
import dbus
from functools import partial

try:
    from pymodbus.client.sync import *
except ImportError:
    from pymodbus.client import *
from pymodbus.register_read_message import ReadHoldingRegistersResponse
import logging
import os
import time
import traceback

from settingsdevice import SettingsDevice
from vedbus import VeDbusService

import __main__
from register import Reg
from utils import *

log = logging.getLogger()


class ModbusDevice(object):
    min_timeout = 0.1

    def __init__(self, modbus, unit, model):
        self.modbus = modbus.get()
        self.unit = unit
        self.model = model
        self.role = None
        self.info = {}
        self.dbus = None
        self.settings = None
        self.err_count = 0
        self.latency = modbus.timeout
        self.need_reinit = False

    def destroy(self):
        self.info = {}
        if self.dbus:
            self.dbus.__del__()
            self.dbus = None
        if self.settings:
            self.settings._settings = None
            self.settings = None
        self.modbus.put()

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return str(self) == str(other)
        if isinstance(other, (str, type(""))):
            return str(self) == other
        return False

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        if isinstance(self.modbus, ModbusTcpClient):
            return "tcp:%s:%d:%d" % (self.modbus.host, self.modbus.port, self.unit)
        elif isinstance(self.modbus, ModbusUdpClient):
            return "udp:%s:%d:%d" % (self.modbus.host, self.modbus.port, self.unit)
        elif isinstance(self.modbus, ModbusSerialClient):
            return "%s:%s:%d:%d" % (
                self.modbus.method,
                os.path.basename(self.modbus.port),
                self.modbus.baudrate,
                self.unit,
            )
        return str(self.modbus)

    def connection(self):
        if isinstance(self.modbus, (ModbusTcpClient, ModbusUdpClient)):
            return "Modbus %s %s" % (
                self.method.upper(),
                self.modbus.socket.getpeername()[0],
            )
        elif isinstance(self.modbus, ModbusSerialClient):
            return "Modbus %s %s:%d" % (
                self.modbus.method.upper(),
                os.path.basename(self.modbus.port),
                self.unit,
            )
        return "Modbus"

    def read_register(self, reg):
        rr = self.modbus.read_holding_registers(reg.base, reg.count, unit=self.unit)

        if not isinstance(rr, ReadHoldingRegistersResponse):
            log.error("Error reading register %#04x: %s", reg.base, rr)
            raise Exception(rr)

        reg.decode(rr.registers)
        return reg.value

    def write_modbus(self, base, val):
        if len(val) == 1:
            self.modbus.write_register(base, val[0], unit=self.unit)
        else:
            self.modbus.write_registers(base, val, unit=self.unit)

    def write_register(self, reg, val):
        reg.value = val
        self.write_modbus(reg.base, reg.encode())

    def read_info_regs(self, d):
        for reg in self.info_regs:
            self.read_register(reg)
            d[reg.name] = reg

    def read_data_regs(self, regs, d):
        now = time.time()

        if all(now - r.time < r.max_age for r in regs):
            return

        start = regs[0].base
        count = regs[-1].base + regs[-1].count - start

        rr = self.modbus.read_holding_registers(start, count, unit=self.unit)

        latency = time.time() - now

        if not isinstance(rr, ReadHoldingRegistersResponse):
            log.error("Error reading registers %#04x-%#04x: %s", start, start + count - 1, rr)
            raise Exception(rr)

        for reg in regs:
            base = reg.base - start
            end = base + reg.count

            if now - reg.time > reg.max_age:
                if reg.decode(rr.registers[base:end]):
                    d[reg.name] = copy(reg) if reg.isvalid() else None
                reg.time = now

        return latency

    def read_info(self):
        if not self.info:
            self.read_info_regs(self.info)

    def init_device_settings(self, dbus):
        if self.settings:
            return

        self.settings_dbus = dbus

        path = "/Settings/Devices/" + self.get_ident()
        role = self.role or self.default_role
        def_inst = "%s:%s" % (self.default_role, self.default_instance)

        SETTINGS = {
            "instance": [path + "/ClassAndVrmInstance", def_inst, 0, 0],
            "customname": [path + "/CustomName", "", 0, 0],
        }

        self.settings = SettingsDevice(dbus, SETTINGS, self.setting_changed)
        self.role, self.devinst = self.get_role_instance()

        if self.role == "pvinverter":
            self.settings.addSettings({"position": [path + "/Position", 0, 0, 2]})

    def setting_changed(self, name, old, new):
        if name == "customname":
            self.dbus["/CustomName"] = new
            return

        if name == "instance":
            role, inst = self.get_role_instance()

            if role != self.role:
                self.role = role
                self.sched_reinit()
                return

            self.dbus["/DeviceInstance"] = inst
            return

        if name == "position":
            if self.role == "pvinverter":
                self.dbus["/Position"] = new
            return

    def get_role_instance(self):
        val = self.settings["instance"].split(":")
        return val[0], int(val[1])

    def reinit(self):
        self.destroy()
        self.init(self.settings_dbus)
        self.need_reinit = False

    def sched_reinit(self):
        self.need_reinit = True

    def get_customname(self):
        return self.settings["customname"]

    def set_customname(self, val):
        self.settings["customname"] = val

    def customname_changed(self, path, val):
        self.set_customname(val)
        return True

    def role_changed(self, path, val):
        if val not in self.allowed_roles:
            return False

        old, inst = self.get_role_instance()
        self.settings["instance"] = "%s:%s" % (val, inst)
        return True

    def position_changed(self, path, val):
        if not 0 <= val <= 2:
            return False

        self.settings["position"] = val
        return True

    def dbus_write_register(self, reg, path, val):
        try:
            val = get_super(Reg, reg)(val)

            if callable(reg.write):
                return reg.write(val)

            if isinstance(reg.write, list):
                if val not in reg.write:
                    return False

            if isinstance(reg.write, tuple):
                if not reg.write[0] <= val <= reg.write[1]:
                    return False

            self.write_register(reg, val)
            return True
        except:
            traceback.print_exc()

        return False

    def dbus_add_register(self, r):
        v = r if r.isvalid() else None
        if r.write:
            cb = partial(self.dbus_write_register, r)
            self.dbus.add_path(r.name, v, writeable=True, onchangecallback=cb)
        else:
            self.dbus.add_path(r.name, v)

    def pack_regs(self, regs):
        rr = []
        for r in regs:
            rr += r if isinstance(r, list) else [r]
        rr.sort(key=lambda r: r.base)

        overhead = 5 + 2  # request + response
        if self.method == "tcp":
            overhead += 2 * (20 + 7)  # TCP + MBAP
        elif self.method == "rtu":
            overhead += 2 * (1 + 2)  # address + crc

        regs = []
        rg = [rr.pop(0)]

        for r in rr:
            end = rg[-1].base + rg[-1].count
            nr = r.base + r.count - rg[0].base
            if nr > 10 or 2 * (r.base - end) > overhead:
                regs.append(rg)
                rg = []

            rg.append(r)

        if rg:
            regs.append(rg)

        return regs

    def init(self, dbus):
        self.device_init()
        self.read_info()
        self.init_device_settings(dbus)

        self.data_regs = self.pack_regs(self.data_regs)
        ident = self.get_ident()

        svcname = "com.victronenergy.%s.%s" % (self.role, ident)
        self.dbus = VeDbusService(svcname, private_bus())

        self.dbus.add_path("/Mgmt/ProcessName", __main__.NAME)
        self.dbus.add_path("/Mgmt/ProcessVersion", __main__.VERSION)
        self.dbus.add_path("/Mgmt/Connection", self.connection())
        self.dbus.add_path("/DeviceInstance", self.devinst)
        self.dbus.add_path("/ProductId", self.productid)
        self.dbus.add_path("/ProductName", self.productname)
        self.dbus.add_path("/Model", self.model)
        self.dbus.add_path("/Connected", 1)
        self.dbus.add_path(
            "/CustomName",
            self.get_customname(),
            writeable=True,
            onchangecallback=self.customname_changed,
        )

        if self.allowed_roles:
            self.dbus.add_path("/AllowedRoles", self.allowed_roles)
            self.dbus.add_path(
                "/Role", self.role, writeable=True, onchangecallback=self.role_changed
            )
        else:
            self.dbus.add_path("/Role", self.role)

        if self.role == "pvinverter":
            self.dbus.add_path(
                "/Position",
                self.settings["position"],
                writeable=True,
                onchangecallback=self.position_changed,
            )

        for p in self.info:
            self.dbus_add_register(self.info[p])

        for r in self.data_regs:
            for rr in r:
                self.dbus_add_register(rr)

        self.latfilt = LatencyFilter(self.latency)
        self.device_init_late()

    def device_init(self):
        pass

    def device_init_late(self):
        pass

    def update(self):
        if self.need_reinit:
            self.reinit()

        latency = []

        with self.dbus as d:
            for r in self.data_regs:
                t = self.read_data_regs(r, d)
                if t:
                    latency.append(t)

        if latency:
            self.latency = self.latfilt.filter(latency)
            self.modbus.timeout = max(self.min_timeout, self.latency * 4)


class LatencyFilter(object):
    def __init__(self, val):
        self.length = 8
        self.pos = 0
        self.val = val
        self.values = [val] * self.length

    def filter(self, values):
        self.values[self.pos] = max(values)
        self.pos += 1
        self.pos &= self.length - 1

        val = max(self.values)

        if val > self.val:
            self.val = 0.25 * self.val + 0.75 * val
        else:
            self.val = 0.75 * self.val + 0.25 * val

        return self.val


class EnergyMeter(ModbusDevice):
    allowed_roles = ["grid", "pvinverter", "genset", "acload"]
    default_role = "grid"
    default_instance = 40


__all__ = [
    "EnergyMeter",
    "ModbusDevice",
]



#!/usr/bin/env python3

import sys
import anyio
from moat.ems.victron.dbus import Dbus
import random


def mon(*a):
    print(a)


async def main():
    async with Dbus() as bus, bus.service("com.victronenergy.battery.test.c") as srv:
        print("Setting up")
        await srv.add_mandatory_paths(
            processname=__file__,
            processversion="0.1",
            connection="Test",
            deviceinstance=1,
            productid=1234,
            productname="Null BMS",
            firmwareversion="0.1",
            hardwareversion=None,
            connected=1,
        )

        V = 26.8
        A = 3.0

        vlo = await srv.add_path(
            "/Info/BatteryLowVoltage",
            0.9 * V,
            gettextcallback=lambda p, v: "{:0.2f}V".format(v),
        )
        vhi = await srv.add_path(
            "/Info/MaxChargeVoltage",
            1.1 * V,
            gettextcallback=lambda p, v: "{:0.2f}V".format(v),
        )
        ich = await srv.add_path(
            "/Info/MaxChargeCurrent",
            A,
            gettextcallback=lambda p, v: "{:0.2f}A".format(v),
        )
        idis = await srv.add_path(
            "/Info/MaxDischargeCurrent",
            A * 1.1,
            gettextcallback=lambda p, v: "{:0.2f}A".format(v),
        )
        ncell = await srv.add_path("/System/NrOfCellsPerBattery", 8)
        non = await srv.add_path("/System/NrOfModulesOnline", 1)
        noff = await srv.add_path("/System/NrOfModulesOffline", 0)
        nbc = await srv.add_path("/System/NrOfModulesBlockingCharge", None)
        nbd = await srv.add_path("/System/NrOfModulesBlockingDischarge", None)
        cap = await srv.add_path("/Capacity", 4.0)
        cap = await srv.add_path("/InstalledCapacity", 5.0)
        cap = await srv.add_path("/ConsumedAmphours", 12.3)

        soc = await srv.add_path("/Soc", 30)
        soh = await srv.add_path("/Soh", 90)
        v0 = await srv.add_path(
            "/Dc/0/Voltage", V, gettextcallback=lambda p, v: "{:2.2f}V".format(v)
        )
        c0 = await srv.add_path(
            "/Dc/0/Current", 0.1, gettextcallback=lambda p, v: "{:2.2f}A".format(v)
        )
        p0 = await srv.add_path(
            "/Dc/0/Power", 0.2, gettextcallback=lambda p, v: "{:0.0f}W".format(v)
        )
        t0 = await srv.add_path("/Dc/0/Temperature", 21.0)
        mv0 = await srv.add_path(
            "/Dc/0/MidVoltage", V / 8, gettextcallback=lambda p, v: "{:0.2f}V".format(v)
        )
        mvd0 = await srv.add_path(
            "/Dc/0/MidVoltageDeviation",
            10.0,
            gettextcallback=lambda p, v: "{:0.1f}%".format(v),
        )

        # battery extras
        minct = await srv.add_path("/System/MinCellTemperature", None)
        maxct = await srv.add_path("/System/MaxCellTemperature", None)
        maxcv = await srv.add_path(
            "/System/MaxCellVoltage",
            None,
            gettextcallback=lambda p, v: "{:0.3f}V".format(v),
        )
        maxcvi = await srv.add_path("/System/MaxVoltageCellId", None)
        mincv = await srv.add_path(
            "/System/MinCellVoltage",
            None,
            gettextcallback=lambda p, v: "{:0.3f}V".format(v),
        )
        mincvi = await srv.add_path("/System/MinVoltageCellId", None)
        hcycles = await srv.add_path("/History/ChargeCycles", None)
        htotalah = await srv.add_path("/History/TotalAhDrawn", None)
        bal = await srv.add_path("/Balancing", None)
        okch = await srv.add_path("/Io/AllowToCharge", 0)
        okdis = await srv.add_path("/Io/AllowToDischarge", 0)
        # xx = await srv.add_path('/SystemSwitch',1)

        # alarms
        allv = await srv.add_path("/Alarms/LowVoltage", None)
        alhv = await srv.add_path("/Alarms/HighVoltage", None)
        allc = await srv.add_path("/Alarms/LowCellVoltage", None)
        alhc = await srv.add_path("/Alarms/HighCellVoltage", None)
        allow = await srv.add_path("/Alarms/LowSoc", None)
        alhch = await srv.add_path("/Alarms/HighChargeCurrent", None)
        alhdis = await srv.add_path("/Alarms/HighDischargeCurrent", None)
        albal = await srv.add_path("/Alarms/CellImbalance", None)
        alfail = await srv.add_path("/Alarms/InternalFailure", None)
        alhct = await srv.add_path("/Alarms/HighChargeTemperature", None)
        allct = await srv.add_path("/Alarms/LowChargeTemperature", None)
        alht = await srv.add_path("/Alarms/HighTemperature", None)
        allt = await srv.add_path("/Alarms/LowTemperature", None)

        await srv.setup_done()

        print("Sending")
        n = 0
        while True:
            n += 1
            await anyio.sleep(n)
            await v0.local_set_value(V * (0.9 + random.random() * 0.2))
            await c0.local_set_value(A * (0.1 + random.random()))
            await p0.local_set_value(v0.value * c0.value)


try:
    anyio.run(main, backend="trio")
except KeyboardInterrupt:
    print("Interrupted.", file=sys.stderr)



from copy import copy
import dbus
from functools import partial
from pymodbus.client import *
from pymodbus.register_read_message import (
    ReadHoldingRegistersResponse,
    ReadInputRegistersResponse,
)
import logging
import os
import time
import traceback
import device

from settingsdevice import SettingsDevice
from vedbus import VeDbusService

import __main__
from register import Reg
from utils import *

log = logging.getLogger()


class ModbusDevice(device.ModbusDevice):
    min_timeout = 0.1

    def read_data_regs(self, regs, d):
        now = time.time()

        if all(now - r.time < r.max_age for r in regs):
            return

        start = regs[0].base
        count = regs[-1].base + regs[-1].count - start

        rr = self.modbus.read_input_registers(start, count, unit=self.unit)

        latency = time.time() - now

        if rr.isError():
            log.error("Error reading registers %#04x-%#04x: %s", start, start + count - 1, rr)
            raise Exception(rr)

        for reg in regs:
            base = reg.base - start
            end = base + reg.count

            if now - reg.time > reg.max_age:
                if reg.decode(rr.registers[base:end]):
                    d[reg.name] = copy(reg) if reg.isvalid() else None
                reg.time = now

        return latency


class CustomName(device.CustomName):
    pass


class EnergyMeter(ModbusDevice):
    allowed_roles = ["grid", "pvinverter", "genset", "acload"]
    default_role = "grid"
    default_instance = 40
    nr_phases = None

    def position_setting_changed(self, service, path, value):
        self.dbus["/Position"] = value["Value"]

    def init_device_settings(self, dbus):
        super().init_device_settings(dbus)

        self.pos_item = None
        if self.role == "pvinverter":
            self.pos_item = self.settings.addSetting(
                self.settings_path + "/Position",
                0,
                0,
                2,
                callback=self.position_setting_changed,
            )

    def device_init_late(self):
        super().device_init_late()

        if self.pos_item is not None:
            self.dbus.add_path(
                "/Position",
                self.pos_item.get_value(),
                writeable=True,
                onchangecallback=self.position_changed,
            )

    def position_changed(self, path, val):
        if not 0 <= val <= 2:
            return False
        self.pos_item.set_value(val)
        return True


__all__ = [
    "EnergyMeter",
    "ModbusDevice",
]



"""
Inverter mode: remote control
"""

from __future__ import annotations

import anyio
import contextlib
import logging

from moat.util import attrdict

from . import InvModeBase

logger = logging.getLogger(__name__)

__all__ = ["InvMode_Remote"]


class InvMode_Remote(InvModeBase):
    """Remote controlled inverter controller."""

    _name = "remote"

    @property
    def power(self):
        "Max power to send to the grid"
        p = self.intf.op.get("power", 0)
        p = max(0, p)
        self.intf.op["power"] = p
        return p

    @property
    def power_ref(self):
        "Reference power, reported to energy marketing provider"
        p = self.intf.op.get("power_ref", 0)
        p = max(0, p)
        self.intf.op["power_ref"] = p
        return p

    @property
    def power_override(self):
        "Inverter power. Set to -1 to disable."
        ip = self.intf.op.get("power_override", None)
        if ip is not None and ip == -1:
            del self.intf.op["power_override"]
            ip = None
        return ip

    @property
    def power_low(self):
        "Power for charging, if the battery is empty"
        p = self.intf.op.get("power_low", 0)
        p = max(0, p)
        self.intf.op["power_low"] = p
        return p

    @property
    def low_grid(self):
        "Do grid zero?"
        p = self.intf.op.get("low_grid", 1)
        p = bool(p)
        self.intf.op["low_grid"] = int(p)
        return p

    @property
    def limit(self):
        "Limit factor. Must be [0…1]."
        p = self.intf.op.get("limit", 1)
        p = max(0, min(1, p))
        self.intf.op["limit"] = p
        return p

    @property
    def mode(self):
        "Operational mode"
        return self.intf.op.get("mode", 3)

    @mode.setter
    def mode(self, val):
        self.intf.op["mode"] = val

    @property
    def soc_low_zero(self):
        "SoC lower? stop the inverter"
        p = self.intf.op.get("soc_low_zero", 0.99)
        p = max(0.05, min(p, self.soc_low - 0.02))
        self.intf.op["soc_low_zero"] = p
        return p

    @property
    def soc_low(self):
        "SoC lower? start grid-only mode"
        p = self.intf.op.get("soc_low", 0.20)
        p = min(max(p, 0.10), 0.80)
        self.intf.op["soc_low"] = p
        return p

    @property
    def soc_low_ok(self):
        "SoC higher? end grid-only mode"
        p = self.intf.op.get("soc_low_ok", 0)
        p = min(max(p, self.soc_low + 0.02), 0.85)
        self.intf.op["soc_low_ok"] = p
        return p

    @property
    def soc_high(self):
        "SoC higher? start feed-out mode"
        p = self.intf.op.get("soc_high", 0.90)
        p = max(min(p, 0.97), self.soc_low + 0.10)
        self.intf.op["soc_high"] = p
        return p

    @property
    def soc_high_ok(self):
        "SoC lower? end feed-out mode"
        p = self.intf.op.get("soc_high_ok", 0.85)
        p = max(min(p, 0.95), self.soc_high - 0.02, self.soc_low + 0.05)
        self.intf.op["soc_high_ok"] = p
        return p

    _doc = dict(
        power=power.__doc__,
        power_low=power_low.__doc__,
        power_ref=power_ref.__doc__,
        power_override=power_override.__doc__,
        limit=limit.__doc__,
        low_grid=low_grid.__doc__,
        soc_low_zero=soc_low_zero.__doc__,
        soc_low=soc_low.__doc__,
        soc_low_ok=soc_low_ok.__doc__,
        soc_high=soc_high.__doc__,
        soc_high_ok=soc_high_ok.__doc__,
        _l="""\
This module implements dynamic control.

When SoC is between @soc_low and @soc_high, the inverter supplies @power to the grid.
(@mode=0)

Above @soc_high it switches to the maximum of @power and solar output. (@mode=3)
It stops doing that when SoC drops below @soc_high_ok.

Below @soc_low the inverter switches to grid-zero if @low_grid is on, else zero.
(mode=2)
Below @soc_low_zero the grid is ignored until @soc_low is reached. (mode=1)
Normal operation is resumed when SoC is higher than @soc_low_ok.

SoC values must be between 0 and 1, though values outside the 0.10 … 0.95 range are
unlikely to work the way you want them to.

External power is constrained by the "limit" value, which must be between 0 and 1.
If DistKV is active, this value is the minimum of the available "limit" sub-entries.

If @power_override is set, its value controls the inverter power directly.
This is intended to shut down the system at night / in low-battery situations.
DO NOT use this setting to feed energy to the grid. In other words, the value
should be negative (-35 per inverter if there's anything on AC OUT).
Set to -1 to clear.
""",
    )

    _limit = None
    _limits: dict = None
    _powers: dict = None
    _power: float = None

    async def run(self):
        "do the work"
        self._limits = {}
        self._powers = {}
        intf = self.intf

        dkv = None
        try:
            async with anyio.create_task_group() as tg:
                evt_l = anyio.Event()
                evt_p = anyio.Event()
                tg.start_soon(self._dkv_mon_l, evt_l)
                tg.start_soon(self._dkv_mon_p, evt_p)
                await evt_l.wait()
                await evt_p.wait()
                tg.start_soon(self._run)

                dkv = await intf.distkv
                if dkv:
                    await dkv.set(intf.distkv_prefix / "solar" / "online", True, idem=True)
        finally:
            if dkv:
                await dkv.set(intf.distkv_prefix / "solar" / "online", False, idem=True)

    async def _dkv_mon_l(self, evt):
        lims = self._limits
        intf = self.intf
        dkv = await intf.distkv
        if dkv is None:
            evt.set()
            return
        async with dkv.watch(intf.distkv_prefix / "solar" / "limit", fetch=True) as mon:
            async for msg in mon:
                if "state" in msg:
                    if msg.state == "uptodate":
                        evt.set()
                        # got them all
                        self._limit = min(lims.values(), default=1)
                else:
                    lim = msg.get("value", None)
                    p = msg.path[-1]
                    logger.info("LIM %s: %f", p, lim)
                    if lim is None:
                        lims.pop(p, None)
                    else:
                        lims[p] = lim
                    if self._limit is not None:
                        self._limit = min(lims.values(), default=1)
                        logger.info("LIM: %f", self._limit)

    async def _dkv_mon_p(self, evt):
        pows = self._powers
        intf = self.intf
        dkv = await intf.distkv
        if dkv is None:
            evt.set()
            return
        async with dkv.watch(intf.distkv_prefix / "solar" / "power", fetch=True) as mon:
            async for msg in mon:
                if "state" in msg:
                    if msg.state == "uptodate":
                        evt.set()
                        # got them all
                        self._power = min(pows.values(), default=0)
                else:
                    val = msg.get("value", None)
                    p = msg.path[-1]
                    logger.info("POW %s: %f", p, val)
                    if val is None:
                        pows.pop(p, None)
                    else:
                        pows[p] = val
                    self._power = min(pows.values(), default=0)
                    logger.info("POW: %f", self._power)

    async def _run(self):
        intf = self.intf
        dkv = await intf.distkv
        state = attrdict(mode=0, limits=self._limits, powers=self._powers)
        intf.set_state("remote", state)

        while True:
            if dkv is None:
                state.limits["manual"] = self._limit = self.limit
                state.powers["manual"] = self._power = self.power
            else:
                if "limit" in self.intf.op:
                    logger.warning("DistKV is active: manual limit is ignored!")
                    del self.intf.op["limit"]
                if "power" in self.intf.op:
                    logger.warning("DistKV is active: power setting is ignored!")
                    del self.intf.op["power"]

            p: float = None
            ip: float = None
            soc = intf.batt_soc
            if soc <= self.soc_low_zero:
                self.mode = 1
            elif self.mode == 1 and soc >= self.soc_low:
                self.mode = 2

            if self.mode != 1 and soc <= self.soc_low:
                self.mode = 2
            elif self.mode in (1, 2) and soc >= self.soc_low_ok:
                self.mode = 0

            if soc >= self.soc_high:
                self.mode = 3
            elif self.mode == 3 and soc <= self.soc_high_ok:
                self.mode = 0

            if self.mode == 1 or (self.mode == 2 and not self.low_grid):
                ip = -self.power_low
            elif self.mode == 2:
                ip = min(intf.solar_p, -intf.p_cons)
            elif self.mode == 3:
                p = max(intf.solar_p + intf.p_cons, self._power)
            else:
                p = self._power
            state.mode = self.mode
            state.p_want = p
            state.ip = ip

            if dkv:
                await dkv.set(intf.distkv_prefix / "solar" / "cur", max(0, -intf.p_grid))
                await dkv.set(intf.distkv_prefix / "solar" / "ref", self.power_ref, idem=True)
            ipn = self.power_override
            if ipn is not None:
                ip = ipn
            if ip is None:
                if dkv:
                    await dkv.set(intf.distkv_prefix / "solar" / "max", p, idem=True)
                exc = self.power_ref - p
                if self._limit is not None:
                    p *= self._limit
                    exc *= self._limit
                state.p_real = p
                state.exc = exc
                with contextlib.suppress(AttributeError):
                    del state.p_i
                ps = intf.calc_grid_p(
                    -p,  # pylint: disable=invalid-unary-operand-type # WAT
                    excess=exc,
                )
            else:
                state.p_i = ip
                with contextlib.suppress(AttributeError):
                    del state.p_real
                with contextlib.suppress(AttributeError):
                    del state.exc
                if dkv:
                    await dkv.set(intf.distkv_prefix / "solar" / "max", 0, idem=True)
                ps = intf.calc_inv_p(ip, excess=0)

            logger.debug("P: %s - IP: %s = %s", p, ip, ps)
            await self.set_inv_ps(ps)
            # already calls "intf.trigger", so we don't have to



# VenusOS module for support of Eastron SDM72D-Modbus v2
# might work also with other Eastron devices > Product code on 0xfc02 (type u16b) to be added into models overview
#
# Community contribution by Thomas Weichenberger
# Version 1.4 - 2022-02-13
#
# Thanks to Victron for their open platform and especially for the support of Matthijs Vader
# For any usage a donation to seashepherd.org with an amount of 5 USD/EUR/GBP or equivalent is expected

import logging
import Eastron_device as device
import probe
from register import *

log = logging.getLogger()


class Reg_f32b(Reg_num):
    coding = (">f", ">2H")
    count = 2
    rtype = float


nr_phases = [0, 1, 3, 3]

phase_configs = [
    "undefined",
    "1P",
    "3P.1",
    "3P.n",
]


class Eastron_SDM72Dv2(device.EnergyMeter):
    productid = 0xB023  # id assigned by Victron Support
    productname = "Eastron SDM72Dv2"
    min_timeout = 0.5

    def __init__(self, *args):
        super(Eastron_SDM72Dv2, self).__init__(*args)

        self.info_regs = [
            Reg_u16(0xFC02, "/HardwareVersion"),
            Reg_u16(0xFC03, "/FirmwareVersion"),
            Reg_f32b(0x000A, "/PhaseConfig", text=phase_configs, write=(0, 3)),
            Reg_u32b(0x0014, "/Serial"),
        ]

    def phase_regs(self, n):
        s = 0x0002 * (n - 1)
        return [
            Reg_f32b(0x0000 + s, "/Ac/L%d/Voltage" % n, 1, "%.1f V"),
            Reg_f32b(0x0006 + s, "/Ac/L%d/Current" % n, 1, "%.1f A"),
            Reg_f32b(0x000C + s, "/Ac/L%d/Power" % n, 1, "%.1f W"),
            Reg_f32b(0x015A + s, "/Ac/L%d/Energy/Forward" % n, 1, "%.1f kWh"),
            Reg_f32b(0x0160 + s, "/Ac/L%d/Energy/Reverse" % n, 1, "%.1f kWh"),
        ]

    def device_init(self):
        self.read_info()

        phases = nr_phases[int(self.info["/PhaseConfig"])]

        regs = [
            Reg_f32b(0x0034, "/Ac/Power", 1, "%.1f W"),
            Reg_f32b(0x0030, "/Ac/Current", 1, "%.1f A"),
            Reg_f32b(0x0046, "/Ac/Frequency", 1, "%.1f Hz"),
            Reg_f32b(0x0048, "/Ac/Energy/Forward", 1, "%.1f kWh"),
            Reg_f32b(0x004A, "/Ac/Energy/Reverse", 1, "%.1f kWh"),
        ]

        for n in range(1, phases + 1):
            regs += self.phase_regs(n)

        self.data_regs = regs

    def get_ident(self):
        return "cg_%s" % self.info["/Serial"]


# identifier to be checked, if register identical on all SDM630 (only first 16 bytes in u16b of 32 bit register 0xfc02)
models = {
    137: {
        "model": "SDM72DMv2",
        "handler": Eastron_SDM72Dv2,
    },
    135: {
        "model": "SDM72Dv2",
        "handler": Eastron_SDM72Dv2,
    },
}

probe.add_handler(probe.ModelRegister(Reg_u16(0xFC02), models, methods=["tcp", "rtu"], units=[1]))



"""
Inverter mode: turn the thing off
"""

from __future__ import annotations

import anyio
import logging

from . import InvModeBase

logger = logging.getLogger(__name__)

__all__ = ["InvMode_None"]


class InvMode_None(InvModeBase):
    "Set the AC output to zero, then do nothing."

    _mode = 0
    _name = "off"

    @property
    def power(self):
        "The power output(+)/input(-) to set"
        return self.intf.op.get("power", 0)

    _doc = dict(
        power=power.__doc__,
        _l="""\
This module sets the inverter power to a specific value,
defaulting to zero, and then does nothing.

Use this to tell the inverter controller to go out of the way
and disable itself temporarily, e.g. while you're testing some
other program.

The power level is from the point of view of the AC side, i.e.
positive = inverter, negaive = charger.
""",
    )

    async def run(self):
        "do the work"
        intf = self.intf

        logger.info("SET inverter ZERO %.0f", self.power)
        for p in intf.p_set_:
            await p.set_value(-self.power / intf.n_phase)
        await anyio.sleep_forever()
